import 'package:bloc_stream/bloc_stream.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod/src/value_provider.dart';
import 'package:meta/meta.dart';

// ignore: implementation_imports
import 'package:riverpod/src/framework.dart';

import 'builders.dart';

mixin _BlocStreamProviderMixin<Bloc extends BlocStream<Value>, Value>
    on ProviderBase<Value> {
  ProviderBase<Bloc> get bloc;

  @override
  ProviderBase<Object?> get providerToRefresh => bloc;

  @override
  void setupOverride(SetupOverride setup) {
    setup(origin: this, override: this);
    setup(origin: bloc, override: bloc);
  }

  Override overrideWithValue(Bloc value) {
    return ProviderOverride((setup) {
      setup(
        origin: bloc,
        override: ValueProvider<Bloc>(value),
      );
      setup(origin: this, override: this);
    });
  }
}

// ignore: subtype_of_sealed_class
@sealed
class _BlocStreamProvider<Bloc extends BlocStream<Object?>>
    extends AlwaysAliveProviderBase<Bloc> {
  _BlocStreamProvider(
    this._create, {
    required String? name,
  }) : super(name == null ? null : '$name.bloc');

  final Create<Bloc, ProviderRefBase> _create;

  @override
  Bloc create(ProviderRefBase ref) {
    final bloc = _create(ref);
    ref.onDispose(bloc.close);
    return bloc;
  }

  @override
  bool updateShouldNotify(Bloc previousState, Bloc newState) {
    return true;
  }

  @override
  ProviderElement<Bloc> createElement() => ProviderElement(this);

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override BlocStreamProvider.bloc');
}

typedef BlocStreamProviderRef<Bloc extends BlocStream<Value>, Value>
    = ProviderRefBase;

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AlwaysAliveProviderBase<Value>
    with _BlocStreamProviderMixin<Bloc, Value> {
  BlocStreamProvider(this._create, {String? name}) : super(name);

  static const family = BlocStreamProviderFamilyBuilder();

  static const autoDispose = AutoDisposeBlocStreamProviderBuilder();

  final Create<Bloc, BlocStreamProviderRef<Bloc, Value>> _create;

  @override
  late final AlwaysAliveProviderBase<Bloc> bloc =
      _BlocStreamProvider<Bloc>(_create, name: name);

  @override
  Value create(ProviderElementBase<Value> ref) {
    final bloc = ref.watch(this.bloc);

    void listener(Value newValue) {
      ref.state = newValue;
    }

    final subscription = bloc.skip(1).listen(listener);
    ref.onDispose(subscription.cancel);

    ref.state = bloc.value;
    return ref.state;
  }

  @override
  bool updateShouldNotify(Value previousState, Value newState) {
    return true;
  }

  Override overrideWithProvider(
    BlocStreamProvider<Bloc, Value> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: bloc, override: provider.bloc);
      setup(origin: this, override: this);
    });
  }

  @override
  ProviderElementBase<Value> createElement() => ProviderElement(this);
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProviderFamily<Bloc extends BlocStream<Value>, Value, Arg>
    extends Family<Value, Arg, BlocStreamProvider<Bloc, Value>> {
  BlocStreamProviderFamily(
    this._create, {
    String? name,
  }) : super(name);

  final FamilyCreate<Bloc, BlocStreamProviderRef<Bloc, Value>, Arg> _create;

  @override
  BlocStreamProvider<Bloc, Value> create(
    Arg argument,
  ) {
    final provider = BlocStreamProvider<Bloc, Value>(
      (ref) => _create(ref, argument),
      name: name,
    );
    registerProvider(provider.bloc, argument);
    return provider;
  }

  /// Overrides the behavior of a family for a part of the application.
  Override overrideWithProvider(
    BlocStreamProvider<Bloc, Value> Function(Arg argument) override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.bloc, override: override(arg).bloc);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.bloc, override: provider.bloc);
  }
}

/// {@macro riverpod.providerrefbase}
typedef AutoDisposeBlocStreamProviderRef<Bloc extends BlocStream<Value>, Value>
    = AutoDisposeProviderRefBase;

class _AutoDisposeBlocProvider<Bloc extends BlocStream<Object?>>
    extends AutoDisposeProviderBase<Bloc> {
  _AutoDisposeBlocProvider(this._create, {required String? name})
      : super(name == null ? null : '$name.bloc');

  final Create<Bloc, AutoDisposeProviderRefBase> _create;

  @override
  Bloc create(AutoDisposeProviderRefBase ref) {
    final bloc = _create(ref);
    ref.onDispose(bloc.close);
    return bloc;
  }

  @override
  bool updateShouldNotify(Bloc previousValue, Bloc newValue) {
    return true;
  }

  @override
  AutoDisposeProviderElement<Bloc> createElement() {
    return AutoDisposeProviderElement(this);
  }

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override BlocStreamProvider.bloc');
}

/// {@macro riverpod.blocstreamprovider}
@sealed
class AutoDisposeBlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AutoDisposeProviderBase<Value>
    with _BlocStreamProviderMixin<Bloc, Value> {
  /// {@macro riverpod.blocstreamprovider}
  AutoDisposeBlocStreamProvider(this._create, {String? name}) : super(name);

  /// {@macro riverpod.family}
  static const family = AutoDisposeBlocStreamProviderFamilyBuilder();

  final Create<Bloc, AutoDisposeBlocStreamProviderRef<Bloc, Value>> _create;

  /// {@template riverpod.blocstreamprovider.bloc}
  /// Obtains the [BlocStream] associated with this [AutoDisposeBlocStreamProvider],
  /// without listening to it.
  ///
  /// Listening to this provider may cause providers/widgets to rebuild in the
  /// event that the [BlocStream] it recreated.
  /// {@endtemplate}
  @override
  late final AutoDisposeProviderBase<Bloc> bloc =
      _AutoDisposeBlocProvider(_create, name: name);

  @override
  Value create(AutoDisposeProviderElementBase<Value> ref) {
    final bloc = ref.watch(this.bloc);

    void listener(Value newValue) {
      ref.state = newValue;
    }

    final subscription = bloc.skip(1).listen(listener);
    ref.onDispose(subscription.cancel);

    ref.state = bloc.value;
    return ref.state;
  }

  @override
  bool updateShouldNotify(Value previousValue, Value newValue) {
    return true;
  }

  /// Overrides the behavior of a provider with a another provider.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    AutoDisposeBlocStreamProvider<Bloc, Value> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: bloc, override: provider.bloc);
      setup(origin: this, override: this);
    });
  }

  @override
  AutoDisposeProviderElementBase<Value> createElement() {
    return AutoDisposeProviderElement(this);
  }
}

/// {@template riverpod.blocstreamprovider.family}
/// A class that allows building a [AutoDisposeBlocStreamProvider] from an external parameter.
/// {@endtemplate}
@sealed
class AutoDisposeBlocStreamProviderFamily<Bloc extends BlocStream<Value>, Value,
        Arg>
    extends Family<Value, Arg, AutoDisposeBlocStreamProvider<Bloc, Value>> {
  /// {@macro riverpod.blocstreamprovider.family}
  AutoDisposeBlocStreamProviderFamily(this._create, {String? name})
      : super(name);

  final FamilyCreate<Bloc, AutoDisposeBlocStreamProviderRef<Bloc, Value>, Arg>
      _create;

  @override
  AutoDisposeBlocStreamProvider<Bloc, Value> create(Arg argument) {
    final provider = AutoDisposeBlocStreamProvider<Bloc, Value>(
      (ref) => _create(ref, argument),
      name: name,
    );

    registerProvider(provider.bloc, argument);

    return provider;
  }

  /// Overrides the behavior of a family for a part of the application.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    AutoDisposeBlocStreamProvider<Bloc, Value> Function(Arg argument) override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.bloc, override: override(arg).bloc);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.bloc, override: provider.bloc);
  }
}
